import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
import matplotlib.pyplot as plt

def read_data(file_name, delimiter=','):
    """ Reads the file containing the data and returns the corresponding tensors """
    data = np.loadtxt(file_name, delimiter=delimiter)
    num_targets = 1
    num_vars = data.shape[1] - num_targets
    N = data.shape[0]

    x = torch.tensor(data[:, :num_vars], dtype=torch.float32)
    d = torch.tensor(data[:, num_vars:], dtype=torch.float32).view(N, 1)
    
    return x, d, N, num_vars, num_targets

def normalization(x):
    """ Normalizes the data by centering and scaling the predictor variables """
    mu = x.mean(0)
    sigma = x.std(0)
    x_norm = (x - mu) / sigma

    return x_norm, mu, sigma

def split_data(x, d, val_prop=0.2, test_prop=0.2):
    """ Splits the initial data into training, validation, and testing subsets """
    assert val_prop + test_prop < 1.0

    N = x.size(0)
    indices = torch.randperm(N)
    num_val = int(N * val_prop)
    num_test = int(N * test_prop)
    num_train = N - num_val - num_test

    x = x[indices]
    d = d[indices]

    x_train = x[:num_train]
    d_train = d[:num_train]

    x_val = x[num_train:num_train + num_val]
    d_val = d[num_train:num_train + num_val]

    x_test = x[num_train + num_val:]
    d_test = d[num_train + num_val:]

    return x_train, d_train, x_val, d_val, x_test, d_test

# Define the neural network class
class NeuralNetwork(nn.Module):
    def __init__(self, layer_dims, activations):
        super(NeuralNetwork, self).__init__()
        layers = []
        for i in range(len(layer_dims) - 1):
            layers.append(nn.Linear(layer_dims[i], layer_dims[i + 1]))
            if activations[i] == 'relu':
                layers.append(nn.ReLU())
            elif activations[i] == 'sigmoid':
                layers.append(nn.Sigmoid())
            elif activations[i] == 'linear':
                pass  # Linear activation is implicit
        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)

def calculate_mse_cost(y, d):
    """ Calculates the MSE loss function """
    return ((y - d) ** 2).mean() / 2

# ===================== Part 1: Data Reading and Normalization =====================
print("Reading data ...")

x, d, N, num_vars, num_targets = read_data("food_truck.txt")
# x, d, N, num_vars, num_targets = read_data("houses.txt")

# Displaying the first 10 examples from the dataset
print("Displaying the first 10 examples from the dataset: ")
for i in range(10):
    print(f"x = {x[i]}, d = {d[i]}")

# Normalizing the variables
print("Normalizing the variables ...")
x, mu, sigma = normalization(x)
dmax = d.max()
d = d / dmax

# Splitting the data
x_train, d_train, x_val, d_val, x_test, d_test = split_data(x, d)

# ===================== Part 2: Training =====================

# Hyperparameters
alpha = 0.001
num_iters = 500
layer_dims = [num_vars, 5, 10, num_targets]
activations = ['relu', 'sigmoid', 'linear']

# Model, loss, and optimizer
model = NeuralNetwork(layer_dims, activations)
criterion = nn.MSELoss()
optimizer = optim.SGD(model.parameters(), lr=alpha)

train_costs = []
val_costs = []

for t in range(num_iters):
    # Training forward pass
    model.train()
    y_train = model(x_train)
    train_loss = criterion(y_train, d_train)

    # Backpropagation
    optimizer.zero_grad()
    train_loss.backward()
    optimizer.step()

    # Validation forward pass
    model.eval()
    with torch.no_grad():
        y_val = model(x_val)
        val_loss = criterion(y_val, d_val)

    train_costs.append(train_loss.item())
    val_costs.append(val_loss.item())

print("Final cost on the training set: ", train_costs[-1])
print("Final cost on the validation set: ", val_costs[-1])

# Plotting the evolution of the cost function
plt.figure()
plt.title("Evolution of the cost function during training")
plt.plot(range(num_iters), train_costs, label="Training")
plt.plot(range(num_iters), val_costs, label="Validation")
plt.legend(loc="upper left")
plt.xlabel("Number of iterations")
plt.ylabel("Cost")
plt.show()

# ===================== Part 3: Evaluation on the Test Set =====================
model.eval()
with torch.no_grad():
    y_test = model(x_test)
    test_loss = criterion(y_test, d_test)

print("Test set cost: ", test_loss.item())
